import importlib
import logging
import os
from typing import Any, Callable

from .context import ValidationContext

logger = logging.getLogger(__name__)

REFTAG_PREFIX = "@"


class RefTagRegistry:
    """Global registry mapping reference tags to their handler functions."""

    _handlers: dict[str, Callable] = {}

    @classmethod
    def register(cls, tag: str, handler: Callable):
        """Register a handler function for a reference tag."""
        if tag in cls._handlers:
            raise ValueError(f"Handler already registered for tag: {tag}")
        cls._handlers[tag] = handler
        logger.debug(f"Registered handler for tag: {tag}")

    @classmethod
    def get_handler(cls, tag: str) -> Callable:
        """Get the handler for a reference tag."""
        if tag not in cls._handlers:
            raise ValueError(f"No handler registered for tag: {tag}")
        return cls._handlers[tag]

    @classmethod
    def clear(cls):
        """Clear all registered handlers."""
        cls._handlers.clear()


def reftag(tag: str):
    """Decorator to register a reference tag handler function."""

    def decorator(handler: Callable):
        RefTagRegistry.register(tag, handler)
        return handler

    return decorator


# Implement built-in reftags
@reftag("import")
def import_tag(path: str, _: ValidationContext) -> Any:
    """Handle @import:module.path.to.thing references."""
    module_path, attr = path.rsplit(".", 1)
    try:
        module = importlib.import_module(module_path)
    except ImportError as e:
        raise ValueError(path) from e
    return getattr(module, attr)


@reftag("value")
def value_tag(path: str, context: ValidationContext) -> Any:
    """Handle @value:path.to.value references."""
    return context.get_nested_value(path)


@reftag("env")
def env_tag(name: str, _: ValidationContext) -> str:
    """Handle @env:VARIABLE_NAME references."""
    try:
        return os.environ[name]
    except KeyError as e:
        raise ValueError(f"Environment variable {name} not found") from e



from cyantic.context import ValidationContext


def test_validation_context():
    """Test the ValidationContext class functionality."""
    test_data = {"outer": {"inner": {"value": 42}}}

    # Test basic context management
    with ValidationContext.root_data(test_data):
        assert ValidationContext.get_root_data() == test_data
        assert ValidationContext.get_nested_value("outer.inner.value") == 42

    # Test context is cleared after exit
    assert ValidationContext.get_root_data() is None

    # Test nested context management
    with ValidationContext.root_data(test_data):
        with ValidationContext.root_data({"other": "data"}):
            # Inner context should not override outer
            assert ValidationContext.get_root_data() == test_data

    # Test error cases
    try:
        ValidationContext.get_nested_value("nonexistent.path")
        assert False, "Should have raised ValueError"
    except ValueError:
        pass

    try:
        with ValidationContext.root_data({"a": 1}):
            ValidationContext.get_nested_value("a.b.c")
        assert False, "Should have raised ValueError"
    except ValueError:
        pass



import os
import random
import statistics
from typing import Sequence, Union, overload
from pydantic import BaseModel, Field, ValidationError

from cyantic import Blueprint, blueprint, CyanticModel


class SimpleModel(CyanticModel):
    """A simple model for testing reftags."""

    name: str


class Tensor(BaseModel):
    """A simple mock tensor class that wraps a list of numbers."""

    data: list[float]

    @classmethod
    def from_list(cls, values: Sequence[float]) -> "Tensor":
        return cls(data=list(values))

    def __len__(self) -> int:
        return len(self.data)

    def mean(self) -> float:
        return statistics.mean(self.data)

    def std(self) -> float:
        return statistics.stdev(self.data)

    @overload
    def __getitem__(self, idx: int) -> float: ...

    @overload
    def __getitem__(self, idx: slice) -> "Tensor": ...

    def __getitem__(self, idx: Union[int, slice]) -> Union[float, "Tensor"]:
        if isinstance(idx, slice):
            return Tensor(data=self.data[idx])
        return self.data[idx]


@blueprint(Tensor)
class NormalTensor(Blueprint[Tensor]):
    """Blueprint for creating tensors from a normal distribution."""

    mean: float
    std_dev: float
    size: int = Field(gt=0, description="Size must be a positive integer")

    def build(self) -> Tensor:
        return Tensor(
            data=[random.gauss(self.mean, self.std_dev) for _ in range(self.size)]
        )


@blueprint(Tensor)
class UniformTensor(Blueprint[Tensor]):
    """Blueprint for creating tensors with values from a uniform distribution."""

    low: float
    high: float
    size: int = Field(gt=0, description="Size must be a positive integer")

    def build(self) -> Tensor:
        return Tensor(
            data=[random.uniform(self.low, self.high) for _ in range(self.size)]
        )


class DataContainer(CyanticModel):
    """Example model using blueprint-enabled tensor."""

    values: Tensor


class NestedConfig(BaseModel):
    """A regular Pydantic model for configuration."""

    name: str
    scale: float


class NestedTensorContainer(CyanticModel):
    """A CyanticModel that will be nested inside another CyanticModel."""

    config: NestedConfig
    tensor: Tensor


class ComplexDataContainer(CyanticModel):
    """A CyanticModel containing both regular fields and nested CyanticModels."""

    name: str
    primary: Tensor
    secondary: NestedTensorContainer


def test_value_reference():
    """Test the @value reference functionality."""
    # Test basic value reference
    data = {
        "stuff": {"tensor": {"mean": 0.0, "std_dev": 1.0, "size": 50}},
        "values": "@value:stuff.tensor",
    }

    model = DataContainer.model_validate(data)
    assert isinstance(model.values, Tensor)
    assert len(model.values) == 50

    # Test nested value reference
    nested_data = {
        "config": {
            "tensors": {
                "primary": {"mean": 0.0, "std_dev": 1.0, "size": 30},
                "secondary": {"low": -1.0, "high": 1.0, "size": 20},
            }
        },
        "name": "foo",
        "primary": "@value:config.tensors.primary",
        "secondary": {
            "config": {"name": "nested", "scale": 2.0},
            "tensor": "@value:config.tensors.secondary",
        },
    }

    model = ComplexDataContainer.model_validate(nested_data)
    assert len(model.primary) == 30
    assert len(model.secondary.tensor) == 20

    # Test invalid path
    invalid_data = {"values": "@value:nonexistent.path"}
    try:
        DataContainer.model_validate(invalid_data)
        assert False, "Should have raised ValueError for invalid path"
    except ValueError:
        pass

    # Test invalid traversal
    invalid_traverse = {
        "nested": {"value": 123},
        "values": "@value:nested.value.deeper",
    }
    try:
        DataContainer.model_validate(invalid_traverse)
        assert False, "Should have raised ValueError for invalid traversal"
    except ValueError:
        pass


def test_import_reference(mocker):
    """Test the @import reference functionality."""
    # Mock module with test object
    mock_module = mocker.Mock()
    mock_module.test_value = Tensor.from_list([1.0, 2.0, 3.0])

    # Setup mock for importlib.import_module
    mock_import = mocker.patch("importlib.import_module")

    def _side_effect(name: str):
        if name == "test.module":
            return mock_module
        raise ImportError(f"No module named '{name}'")

    mock_import.side_effect = _side_effect

    # Test successful import
    data = {"values": "@import:test.module.test_value"}
    model = DataContainer.model_validate(data)
    assert isinstance(model.values, Tensor)
    assert model.values.data == [1.0, 2.0, 3.0]

    # Test invalid module
    data = {"values": "@import:nonexistent.module.value"}
    try:
        DataContainer.model_validate(data)
        assert False, "Should have raised ValueError for invalid import"
    except ValueError as e:
        pass

    # Test invalid reference format
    data = {"values": "@invalid:something"}
    try:
        DataContainer.model_validate(data)
        assert False, "Should have raised ValueError for invalid reference type"
    except ValueError as e:
        pass


def test_env_reference():
    """Test the @env reference functionality."""
    # Test successful env var reference
    os.environ["TEST_VAR"] = "test_value"
    data = {"name": "@env:TEST_VAR"}
    model = SimpleModel.model_validate(data)
    assert model.name == "test_value"

    # Test missing env var raises ValidationError
    data = {"name": "@env:NONEXISTENT_VAR"}
    try:
        SimpleModel.model_validate(data)
        assert False, "Should have raised ValidationError"
    except ValidationError as e:
        assert "Environment variable NONEXISTENT_VAR not found" in str(e)



import random
import statistics
from typing import Sequence, Union, overload

from pydantic import BaseModel, Field

from cyantic import Blueprint, blueprint, CyanticModel


class SimpleModel(CyanticModel):
    """A simple model for testing reftags."""

    name: str


class CounterModel(CyanticModel):
    """A model for testing stateful reftags."""

    first: int
    second: int
    third: int


class Tensor(BaseModel):
    """A simple mock tensor class that wraps a list of numbers.

    We use this because we don't want to take torch as a dependency.
    """

    data: list[float]

    @classmethod
    def from_list(cls, values: Sequence[float]) -> "Tensor":
        return cls(data=list(values))

    def __len__(self) -> int:
        return len(self.data)

    def mean(self) -> float:
        return statistics.mean(self.data)

    def std(self) -> float:
        return statistics.stdev(self.data)

    @overload
    def __getitem__(self, idx: int) -> float: ...

    @overload
    def __getitem__(self, idx: slice) -> "Tensor": ...

    def __getitem__(self, idx: Union[int, slice]) -> Union[float, "Tensor"]:
        if isinstance(idx, slice):
            return Tensor(data=self.data[idx])
        return self.data[idx]


@blueprint(Tensor)
class NormalTensor(Blueprint[Tensor]):
    """Blueprint for creating tensors from a normal distribution."""

    mean: float
    std_dev: float
    size: int = Field(gt=0, description="Size must be a positive integer")

    def build(self) -> Tensor:
        return Tensor(
            data=[random.gauss(self.mean, self.std_dev) for _ in range(self.size)]
        )


@blueprint(Tensor)
class UniformTensor(Blueprint[Tensor]):
    """Blueprint for creating tensors with values from a uniform distribution."""

    low: float
    high: float
    size: int = Field(gt=0, description="Size must be a positive integer")

    def build(self) -> Tensor:
        return Tensor(
            data=[random.uniform(self.low, self.high) for _ in range(self.size)]
        )


class DataContainer(CyanticModel):
    """Example model using cast-enabled tensor."""

    values: Tensor


def test_cast_build():
    """Test the cast building functionality."""
    # Test direct tensor assignment
    direct_tensor = Tensor.from_list([1.0, 2.0, 3.0])
    model1 = DataContainer(values=direct_tensor)
    assert len(model1.values) == 3
    assert list(model1.values.data) == [1.0, 2.0, 3.0]

    # Test normal distribution cast - verify only size and type
    cast_dict = {"values": {"mean": 0.0, "std_dev": 1.0, "size": 10}}
    model2 = DataContainer.model_validate(cast_dict)
    assert len(model2.values) == 10
    assert isinstance(model2.values, Tensor)
    assert all(isinstance(x, float) for x in model2.values.data)

    # Test validation error for missing required fields
    try:
        DataContainer.model_validate({"values": {"mean": 0.0}})
        assert False, "Should have raised ValueError"
    except ValueError:
        pass


class TensorList(CyanticModel):
    """A model containing a list of tensors."""

    tensors: list[Tensor]


def test_list_of_castmodels():
    """Test handling lists of fields that use casts."""
    # Test list of normal distributions - verify sizes and types only
    normal_list_data = {
        "tensors": [
            {"mean": 0.0, "std_dev": 1.0, "size": 10},
            {"mean": 5.0, "std_dev": 2.0, "size": 20},
            {"mean": -5.0, "std_dev": 0.5, "size": 30},
        ]
    }

    model = TensorList.model_validate(normal_list_data)
    assert len(model.tensors) == 3
    assert all(isinstance(t, Tensor) for t in model.tensors)
    assert len(model.tensors[0]) == 10
    assert len(model.tensors[1]) == 20
    assert len(model.tensors[2]) == 30
    assert all(isinstance(x, float) for t in model.tensors for x in t.data)

    # Test mixed normal and uniform distributions - verify types and sizes
    mixed_list_data = {
        "tensors": [
            {"mean": 0.0, "std_dev": 1.0, "size": 50},  # Normal
            {"low": 0.0, "high": 1.0, "size": 50},  # Uniform
            {"mean": 2.0, "std_dev": 0.5, "size": 50},  # Normal
        ]
    }

    model = TensorList.model_validate(mixed_list_data)
    assert len(model.tensors) == 3
    assert all(isinstance(t, Tensor) for t in model.tensors)
    assert all(len(t) == 50 for t in model.tensors)
    assert all(isinstance(x, float) for t in model.tensors for x in t.data)

    # Test validation with invalid items in list
    invalid_list_data = {
        "tensors": [
            {"mean": 0.0, "std_dev": 1.0, "size": 10},
            {"mean": 0.0, "size": 20},  # Missing std_dev
            {"low": 0.0, "high": 1.0, "size": 30},
        ]
    }

    try:
        TensorList.model_validate(invalid_list_data)
        assert False, "Should have raised ValueError for invalid tensor spec"
    except ValueError:
        pass

    # Test uniform distribution cast - verify size and types
    model3 = DataContainer.model_validate(
        {"values": {"low": -1.0, "high": 1.0, "size": 50}}
    )
    assert len(model3.values) == 50
    assert isinstance(model3.values, Tensor)
    assert all(isinstance(x, float) for x in model3.values.data)


class NestedConfig(BaseModel):
    """A regular Pydantic model for configuration."""

    name: str
    scale: float


class NestedTensorContainer(CyanticModel):
    """A CyanticModel that will be nested inside another CyanticModel."""

    config: NestedConfig
    tensor: Tensor


class ComplexDataContainer(CyanticModel):
    """A CyanticModel containing both regular fields and nested CyanticModels."""

    name: str
    primary: Tensor
    secondary: NestedTensorContainer


def test_nested_blueprint_models():
    """Test nested CyanticModels with mixed BaseModel types."""
    # Test nested structure with both normal and blueprint fields

    nested_data = {
        "name": "test_complex",
        "primary": {"mean": 0.0, "std_dev": 1.0, "size": 50},
        "secondary": {
            "config": {"name": "nested", "scale": 2.0},
            "tensor": {"low": -1.0, "high": 1.0, "size": 30},
        },
    }

    model = ComplexDataContainer.model_validate(nested_data)

    # Verify top level fields
    assert model.name == "test_complex"
    assert len(model.primary) == 50

    # Verify nested structure
    assert model.secondary.config.name == "nested"
    assert model.secondary.config.scale == 2.0
    assert len(model.secondary.tensor) == 30


def test_mixed_model_validation():
    """Test validation behavior with mixed model types."""
    # Test validation with missing nested fields
    invalid_data = {
        "name": "test_invalid",
        "primary": {"mean": 0.0, "std_dev": 1.0, "size": 50},
        "secondary": {
            "config": {"name": "nested"},  # Missing scale
            "tensor": {"low": -1.0, "high": 1.0, "size": 30},
        },
    }

    try:
        ComplexDataContainer.model_validate(invalid_data)
        assert False, "Should have raised ValueError for missing scale"
    except ValueError:
        pass

    # Test validation with invalid parameters
    invalid_data = {
        "name": "test_invalid",
        "primary": {"mean": 0.0, "std_dev": 1.0, "size": -50},  # Invalid size
        "secondary": {
            "config": {"name": "nested", "scale": 2.0},
            "tensor": {"low": -1.0, "high": 1.0, "size": 30},
        },
    }

    try:
        ComplexDataContainer.model_validate(invalid_data)
        assert False, "Should have raised ValueError for negative size"
    except ValueError:
        pass


def test_blueprint_type_inference():
    """Test that we correctly infer and apply different blueprints."""
    # Test that both NormalTensor and UniformTensor blueprints work in the same model
    mixed_data = {
        "name": "test_mixed",
        "primary": {"mean": 0.0, "std_dev": 1.0, "size": 100},
        "secondary": {
            "config": {"name": "nested", "scale": 2.0},
            "tensor": {"low": 0.0, "high": 1.0, "size": 100},
        },
    }

    model = ComplexDataContainer.model_validate(mixed_data)

    # Verify that different blueprints were correctly applied
    assert len(model.primary) == 100
    assert len(model.secondary.tensor) == 100

    # Verify types and sizes only
    assert isinstance(model.primary, Tensor)
    assert isinstance(model.secondary.tensor, Tensor)
    assert len(model.primary) == 100
    assert len(model.secondary.tensor) == 100
    assert all(isinstance(x, float) for x in model.primary.data)
    assert all(isinstance(x, float) for x in model.secondary.tensor.data)


class TensorWrapper(CyanticModel):
    """A simple CyanticModel that will be nested in a BaseModel."""

    tensor: Tensor


class ConfigWithTensor(BaseModel):
    """A BaseModel that contains a CyanticModel."""

    name: str
    description: str | None = None
    data: TensorWrapper


def test_cyanticmodel_in_basemodel():
    """Test a CyanticModel nested inside a regular Pydantic BaseModel."""
    config_data = {
        "name": "test_config",
        "description": "Testing CyanticModel inside BaseModel",
        "data": {"tensor": {"mean": 0.0, "std_dev": 1.0, "size": 50}},
    }

    model = ConfigWithTensor.model_validate(config_data)

    # Verify BaseModel fields
    assert model.name == "test_config"
    assert model.description == "Testing CyanticModel inside BaseModel"

    # Verify the nested CyanticModel types and sizes
    assert isinstance(model.data.tensor, Tensor)
    assert len(model.data.tensor) == 50
    assert all(isinstance(x, float) for x in model.data.tensor.data)

    # Test validation with invalid nested blueprint
    invalid_data = {
        "name": "test_invalid",
        "data": {
            "tensor": {
                "mean": 0.0,
                "std_dev": 1.0,
                "size": -10,  # Invalid negative size
            }
        },
    }

    try:
        ConfigWithTensor.model_validate(invalid_data)
        assert False, "Should have raised ValueError for negative size"
    except ValueError:
        pass



from cyantic.core import blueprint, Blueprint, CyanticModel
from cyantic.context import ValidationContext
from cyantic.reftag import reftag, RefTagRegistry

__all__ = [
    "blueprint",
    "Blueprint",
    "CyanticModel",
    "reftag",
    "ValidationContext",
    "RefTagRegistry",
]
